/* Simple lock manager without queues but busy loops
 
   The lock directory is expected to have two subdirectories:
   - clients       has a file for each client that is named after
                   its id and has its id as content
   - locks         locks are held by hard-linking the client files
                   out of the clients directory to a file that
                   is named after the lock name
*/

#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <afblib/inbuf.h>
#include <afblib/outbuf.h>
#include "lockmanager.h"

/* read the id out of a lock file */
static bool read_name(inbuf* ibuf, stralloc* name) {
   int ch;
   if (!stralloc_copys(name, "")) return false;
   while ((ch = inbuf_getchar(ibuf)) >= 0 && ch != '\n') {
      if (!stralloc_readyplus(name, 1)) return false;
      name->s[name->len++] = ch;
   }
   return name->len > 0;
}

static bool create_path(stralloc* path,
      lockset* set, char* subdir, char* entry) {
   if (!stralloc_copys(path, set->dirname)) return false;
   if (!stralloc_cats(path, "/")) return false;
   if (!stralloc_cats(path, subdir)) return false;
   if (!stralloc_cats(path, "/")) return false;
   if (!stralloc_cats(path, entry)) return false;
   if (!stralloc_0(path)) return false;
   return true;
}

/* initialize lock set */
bool lm_init(lockset* set, char* dirname, char* myname) {
   if (!strhash_alloc(&set->locks, 4)) return false;
   set->dirname = strdup(dirname);
   if (set->dirname == 0) return false;
   set->myname = strdup(myname);
   if (set->myname == 0) return false;
   set->myfile = (stralloc) {0};

   if (!create_path(&set->myfile, set, "clients", myname)) return false;
   int fd = open(set->myfile.s, O_WRONLY|O_CREAT|O_EXCL, 0444);
   if (fd < 0) return false;
   outbuf obuf = {fd};
   bool ok =
      outbuf_write(&obuf,
         set->myname, strlen(set->myname)) == strlen(set->myname) &&
      outbuf_putchar(&obuf, '\n') == '\n' &&
      outbuf_flush(&obuf);
   outbuf_free(&obuf); close(fd);
   if (!ok) {
      unlink(set->myfile.s); return false;
   }
   return true;
}

/* release all locks associated with set and allocated storage */
void lm_free(lockset* set) {
   if (strhash_start(&set->locks)) {
      char* lockname;
      while (strhash_next(&set->locks, &lockname)) {
         stralloc path = {0};
         if (create_path(&path, set, "locks", lockname)) {
            unlink(path.s);
            stralloc_free(&path);
         }
      }
   }
   unlink(set->myfile.s);
   stralloc_free(&set->myfile);
   free(set->dirname); set->dirname = 0;
   free(set->myname); set->myname = 0;
   strhash_free(&set->locks);
}

/* check status of the given lock and return
   the name of the holder in holder if its held
   and an empty string if the lock is free */
bool lm_stat(lockset* set, char* lockname, stralloc* holder) {
   if (!stralloc_copys(holder, "")) return false;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return false;
   int fd = open(path.s, O_RDONLY);
   stralloc_free(&path);
   if (fd < 0) return true;
   inbuf ibuf = {fd};
   bool ok = read_name(&ibuf, holder);
   close(fd); inbuf_free(&ibuf);
   return ok;
}

/* attempt to lock `lockname' */
bool lm_lock(lockset* set, char* lockname) {
   if (!strhash_add(&set->locks, lockname, 0)) return false;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return false;
   bool ok;
   for(;;) {
      errno = 0;
      if (link(set->myfile.s, path.s) == 0) {
         ok = true; break;
      }
      if (errno != EEXIST) {
         ok = false; break;
      }
      sleep(1);
   }
   stralloc_free(&path);
   if (!ok) strhash_remove(&set->locks, lockname);
   return ok;
}

/* attempt to lock `lockname' but do not block */
bool lm_nonblocking_lock(lockset* set, char* lockname) {
   if (!strhash_add(&set->locks, lockname, 0)) return false;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return false;
   bool ok = link(set->myfile.s, path.s) == 0;
   if (!ok) strhash_remove(&set->locks, lockname);
   stralloc_free(&path);
   return ok;
}

/* release `lockname' */
bool lm_release(lockset* set, char* lockname) {
   if (!strhash_remove(&set->locks, lockname)) return false;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return false;
   bool ok = unlink(path.s) == 0;
   stralloc_free(&path);
   return ok;
}
